#include <vector>
#include <algorithm>

#include <cstdio>
#include <cstring>
#include <cstdlib>

#include "matrix.h"

extern "C" {
#include "mmio.h"
}

/*
 * Load matrix market file
 */

struct Element
{
  int row, col;
  double val;

  Element(int row, int col, double val) : row(row), col(col), val(val) { }
};

static bool
element_sort_func(const struct Element a, const struct Element b)
{
  if (a.row < b.row)
    return true;

  if (a.row == b.row && a.col < b.col)
    return true;

  return false;
}

static bool
read_matrix_market(const char *filename,
                   std::vector<Element> &elements,
                   int &n_rows, int &n_cols)
{
  FILE *fh = fopen(filename, "r");
  if (!fh)
    {
      perror("fopen");
      return false;
    }

  printf("Reading matrix '%s'...\n", filename);

  MM_typecode matcode;
  if (mm_read_banner(fh, &matcode) != 0)
    {
      fprintf(stderr, "mm_read_banner failed\n");
      return false;
    }

  int M, N, nz;
  int ret_code;
  ret_code = mm_read_mtx_crd_size(fh, &M, &N, &nz);
  if (ret_code != 0)
    {
      fprintf(stderr, "mm_read_mtx_crd_size failed\n");
      return false;
    }

  n_rows = M;
  n_cols = N;

  for (int i = 0; i < nz; i++)
    {
      int row, col;
      double val;

      if (mm_is_pattern(matcode))
        {
          fscanf(fh, "%d %d\n", &row, &col);
          val = 1.0;
        }
      else
        fscanf(fh, "%d %d %lg\n", &row, &col, &val);

      row--; /* adjust from 1-based to 0-based */
      col--;

      elements.push_back(Element(row, col, val));
      if (mm_is_symmetric(matcode) && row != col)
        elements.push_back(Element(col, row, val));
    }

  std::sort(elements.begin(), elements.end(), element_sort_func);

  return true;
}

/*
 * Transfer matrix elements into Compressed Row Storage structure
 */

Matrix *
matrix_new(const int n_rows, const int n_cols, const int nnz)
{
  Matrix *M = (Matrix *)malloc(sizeof(Matrix));

  M->nnz = nnz;
  M->n_rows = n_rows;
  M->n_cols = n_cols;

  M->val = (double *)malloc(sizeof(double) * nnz);
  M->col_ind = (int *)malloc(sizeof(int) * nnz);

  M->row_ptr_begin = (int *)malloc(sizeof(int) * n_rows);
  M->row_ptr_end = (int *)malloc(sizeof(int) * n_rows);

  return M;
}

void
matrix_free(Matrix *M)
{
  if (!M)
    return;

  free(M->val);
  free(M->col_ind);
  free(M->row_ptr_begin);
  free(M->row_ptr_end);
  free(M);
}

static void
matrix_load_elements(Matrix *A, const std::vector<Element> &elements)
{
  int current_val = 0;
  int current_row = 0;
  A->row_ptr_begin[0] = 0;

  for (std::vector<Element>::const_iterator it = elements.begin();
       it != elements.end(); ++it)
    {
      if (it->row != current_row)
        {
          if (current_row + 1 != it->row)
            {
              fprintf(stderr, "Row skipping not implemented.\n");
              abort();
            }

          A->row_ptr_end[current_row] = current_val - 1;
          current_row++;
          A->row_ptr_begin[current_row] = current_val;
        }

      A->val[current_val] = it->val;
      A->col_ind[current_val] = it->col;
      current_val++;
    }

  A->row_ptr_end[current_row] = current_val - 1;
}

void
matrix_dump_nonzeros(Matrix *A)
{
  for (int row = 0; row < A->n_rows; row++)
    {
      for (int idx = A->row_ptr_begin[row]; idx <= A->row_ptr_end[row]; ++idx)
        {
          printf("%d %d %f\n", row, A->col_ind[idx], A->val[idx]);
        }
    }
}

Matrix *
matrix_load_matrix_market(const char *filename)
{
  int n_rows, n_cols;
  std::vector<Element> elements;

  if (!read_matrix_market(filename, elements, n_rows, n_cols))
    return NULL;

  Matrix *A = matrix_new(n_rows, n_cols, elements.size());
  matrix_load_elements(A, elements);

  printf("import ok: %d x %d matrix, %d nnz\n",
         A->n_rows, A->n_cols, A->nnz);

  return A;
}
